import numpy as np
from math import factorial
import random
from tqdm import tqdm
import matplotlib.pyplot as plt


class ShapleyValueCalculator:
    def __init__(self, model, X, y, num_samples=1000, random_state=42):
        """
        Initialize Shapley Value Calculator
        
        Args:
            model: Trained model with predict method
            X: numpy array or list - input data shape (n_samples, n_features)
            y: numpy array or list - target values
            num_samples: int - number of samples for Shapley value estimation
            random_state: int - seed for reproducibility
        """
        self.model = model
        self.X = np.array(X)
        self.y = np.array(y)
        self.num_samples = num_samples
        self.n_features = self.X.shape[1]
        self.baseline = self._get_baseline_prediction()
        self.random_state = random_state
        # Set seed
        random.seed(self.random_state)
        np.random.seed(self.random_state)

    def _get_baseline_prediction(self):
        """Get baseline prediction on zero features"""
        X_baseline = np.zeros((1, self.n_features))
        return self.model.predict(X_baseline)[0]

    def _calculate_weights(self):
        """Pre-calculate weights for all coalition sizes"""
        n = self.n_features
        weights = np.zeros(n + 1)
        for size in range(1, n):
            weights[size] = factorial(size - 1) * factorial(n - size) / factorial(n)
        return weights

    def _get_predictions_batch(self, x, coalitions):
        """Get model predictions for a batch of coalitions"""
        X_modified = np.zeros((len(coalitions), self.n_features))
        for i, coalition in enumerate(coalitions):
            X_modified[i, coalition] = x[coalition]
        return self.model.predict(X_modified)

    def get_shapley_values(self, sample_idx=None, normalize=False):
        """
        Calculate Shapley values for a given example or averaged over all examples
        
        Args:
            sample_idx: index of the example to calculate or None to average over all examples
            normalize: bool, whether to normalize Shapley values to sum to 1
        """
        if sample_idx is not None:
            values = self._compute_shapley_for_sample(self.X[sample_idx])
        else:
            print("Computing Shapley values for all samples...")
            values = np.mean([self._compute_shapley_for_sample(x) for x in tqdm(self.X)], axis=0)
        
        if normalize:
            # Handle case when there are negative values
            abs_values = np.abs(values)
            total = np.sum(abs_values)
            if total != 0:
                # Normalize while preserving signs and ensuring exact sum of 1
                signs = np.sign(values)
                normalized = abs_values / total
                values = signs * normalized
                
                # Adjust to ensure exact sum of 1
                values = values / np.sum(np.abs(values))
                
        return values

    def _compute_shapley_for_sample(self, x):
        """Calculate Shapley values for a single example"""
        # Set seed before each calculation
        random.seed(self.random_state)
        
        shapley_values = np.zeros(self.n_features)
        weights = self._calculate_weights()
        batch_size = 32  # Batch size for predictions
        
        # Generate all coalitions in advance
        for i in range(self.n_features):
            coalitions_with_i = []
            coalitions_without_i = []
            n_samples_per_size = max(1, self.num_samples // (self.n_features - 1))
            
            for size in range(1, self.n_features):
                for _ in range(n_samples_per_size):
                    available_features = list(set(range(self.n_features)) - {i})
                    coalition = random.sample(available_features, size - 1)
                    coalitions_without_i.append(coalition)
                    coalitions_with_i.append(coalition + [i])
            
            # Batch processing
            for j in range(0, len(coalitions_with_i), batch_size):
                batch_with = coalitions_with_i[j:j + batch_size]
                batch_without = coalitions_without_i[j:j + batch_size]
                
                pred_with_i = self._get_predictions_batch(x, batch_with)
                pred_without_i = self._get_predictions_batch(x, batch_without)
                
                # Calculate marginal contributions for the batch
                marginal_contributions = pred_with_i - pred_without_i
                
                # Update Shapley values with weights
                for k, (size, margin) in enumerate(zip([len(c) for c in batch_with], marginal_contributions)):
                    shapley_values[i] += weights[size] * margin / n_samples_per_size
        
        return shapley_values

    def plot_shapley_values(self, feature_names=None, sample_idx=None, normalize=True, title=None):
        """
        Visualize Shapley values as a horizontal bar chart
        
        Args:
            feature_names: list of feature names
            sample_idx: index of a specific example or None for averaged values
            normalize: whether to normalize Shapley values
            title: title of the chart
        """
        values = self.get_shapley_values(sample_idx=sample_idx, normalize=normalize)
        
        if feature_names is None:
            feature_names = [f"Feature {i}" for i in range(self.n_features)]
            
        # Create color scheme
        colors = ['darkred' if x > 0 else 'darkblue' for x in values]
        alphas = np.abs(values) / max(np.abs(values))  # Normalize for transparency
        
        # Create plot
        plt.figure(figsize=(10, 6))
        y_pos = np.arange(len(feature_names))
        
        bars = plt.barh(y_pos, values, align='center', alpha=0.8)
        
        # Set colors with gradient
        for bar, color, alpha in zip(bars, colors, alphas):
            bar.set_color(color)
            bar.set_alpha(alpha)
        
        plt.yticks(y_pos, feature_names)
        plt.xlabel('Shapley Value')
        plt.title(title or 'Feature Importance (Shapley Values)')
        
        # Add vertical line at zero
        plt.axvline(x=0, color='black', linestyle='-', linewidth=0.5)
        
        # Add values at the ends of bars
        for i, v in enumerate(values):
            plt.text(v + np.sign(v) * 0.01, i, f'{v:.3f}', 
                    ha='left' if v > 0 else 'right', 
                    va='center')
        
        plt.tight_layout()
        return plt.gcf()